var http = require('http');
var Url = require("url");
var querystring = require('querystring');
var path = require("path");
var fs = require("fs");
var utils = require("node-pearls");
var dbPath = path.join(process.env.HOME || __dirname, ".dev-static-server:db");

utils.mkdirs.sync(dbPath);

function parseQueryString(data){
	try{
		data = querystring.parse(data);
		for(var name in data){
			try{
				if(data[name] === "undefined"){
					delete data[name];
				}else{
					data[name] = JSON.parse(data[name]);
				}
			}catch(e){}
		}
		return data;
	}catch(e){
		throw e;
	}
}

function parseFunction(columns){
	var result = columns.trim().match(/^(first|last|min|max|sum|count|avg)\(([^\)]+)\)$/);
	if(result){
		return {
			func: result[1],
			columns: result[2]
		};
	}else{
		return {
			func: null,
			columns: columns
		};
	}
}

function readRequestData(request){
	return new Promise(function(resolve, reject){
		var data = "";
		request.setEncoding('utf8');
		request.on('data', function(chunk){
			data += chunk;
		});
		request.on('end', function(){
			try{
				data = parseQueryString(data);
				resolve(data);
			}catch(e){
				reject(e);
			}
		});
	});
}

function filterWhere(data, where){
	if(!where){
		return data;
	}

	var names = Object.keys(where);

	return data.filter(function(item){
		return names.every(function(name){
			return item[name] === where[name];
		});
	});
}

function filterColumns(data, columns){
	if(columns.trim() === "*"){
		return data;
	}

	columns = columns.split(",").map(function(col){
		return col.trim();
	});

	return data.map(function(item){
		var newData = {};
		columns.forEach(function(name){
			newData[name] = item[name];
		});
		return newData;
	});
}

function log(type, tableName, info, result){
	switch(type){
		case "select":
			console.log(`select ${info.columns} from ${tableName} where ${querystring.stringify(info.where)}`);
			console.log(result);
			break;
		case "insert":
			console.log(`insert into ${tableName} values ${JSON.stringify(info.values)}`);
			console.log(`index: ${result}`);
			break;
		case "update":
			console.log(`update ${tableName} set ${querystring.stringify(info.sets)} where ${querystring.stringify(info.where)}`);
			console.log(`update count: ${result}`);
			break;
		case "delete":
			console.log(`delete from ${tableName} where ${querystring.stringify(info.where)}`);
			console.log(`delete count: ${result}`);
			break;
		case "create table":
			console.log(`create table ${info.tableName} ${JSON.stringify(info.columns)}`);
			console.log(result);
			break;
	}
}

var cache = {};

var server = http.createServer(function (req, res) {
	var url = Url.parse(req.url);
	var tableName = url.pathname.replace(/^\//, "");
	var info = parseQueryString(url.query);
	
	if(tableName){
		var tableFile = path.join(dbPath, tableName + ".json");

		if(!cache[tableName]){
			var tableInfo = JSON.parse(fs.readFileSync(tableFile, "utf8"));
			cache[tableName] = {
				save: function(){
					fs.writeFileSync(tableFile, JSON.stringify(tableInfo));
				}
			};
			Object.defineProperties(cache[tableName], {
				info: {
					get: function(){
						return tableInfo.info;
					}
				},
				data: {
					get: function(){
						return tableInfo.data;
					},
					set: function(data){
						tableInfo.data = data;
					}
				}
			});
		}

		var table = cache[tableName];

		switch(req.method){
			case "GET":
				// select columns where
				res.writeHead(200, {
					'Content-Type': 'application/json'
				});
				var result = filterWhere(table.data, info.where);
				var filter = parseFunction(info.columns);
				result = filterColumns(result, filter.columns);
				if(filter.func){
					if(result.length){
						switch(filter.func){
							case "first":
								result = result[0];
								break;
							case "last":
								result = result[result.length - 1];
								break;
							case "min":
								var min = Infinity;
								result.forEach(function(item){
									var value = item[filter.columns];
									if(typeof value === "number"){
										if(value < min){
											min = value;
										}
									}else{
										console.log(`非number类型的字段${filter.columns}不能进行min运算`);
									}
								});
								result = min;
								break;
							case "max":
								var max = -Infinity;
								result.forEach(function(item){
									var value = item[filter.columns];
									if(typeof value === "number"){
										if(value > max){
											max = value;
										}
									}else{
										console.log(`非number类型的字段${filter.columns}不能进行max运算`);
									}
								});
								result = max;
								break;
							case "sum":
								var sum = 0;
								result.forEach(function(item){
									var value = item[filter.columns];
									if(typeof value === "number"){
										sum += value;
									}else{
										console.log(`非number类型的字段${filter.columns}不能进行sum运算`);
									}
								});
								result = sum;
								break;
							case "count":
								result = result.length;
								break;
							case "avg":
								var total = 0;
								result.forEach(function(item){
									var value = item[filter.columns];
									if(typeof value === "number"){
										total += value;
									}else{
										console.log(`非number类型的字段${filter.columns}不能进行avg运算`);
									}
								});
								result = total / result.length;
								break;
						}
					}else{
						result = null;
					}
				}
				res.write(JSON.stringify(result) || "");
				res.end();
				log("select", tableName, info, result);
				break;
			case "POST":
				// insert values
				readRequestData(req).then(function(info){
					res.writeHead(201, {
						'Content-Type': 'application/json'
					});
					var index = table.data.push(info.values);
					table.save();
					res.write(JSON.stringify(index));
					res.end();
					log("insert", tableName, info, index);
				});
				break;
			case "PUT":
				// update sets where
				readRequestData(req).then(function(info){
					res.writeHead(200, {
						'Content-Type': 'application/json'
					});
					var result = filterWhere(table.data, info.where);
					var setNames = Object.keys(info.sets);
					result.forEach(function(item){
						setNames.forEach(function(name){
							item[name] = info.sets[name];
						});
					});
					table.save();
					res.write(JSON.stringify(result.length));
					res.end();
					log("update", tableName, info, result.length);
				});
				break;
			case "DELETE":
				// delete where
				res.writeHead(200, {
					'Content-Type': 'application/json'
				});
				var whereNames = Object.keys(where);
				var newData = table.data.filter(function(item){
					return whereNames.some(function(name){
						item[name] !== where[name];
					});
				});
				var changeCount = table.data.length - newData.length;
				if(changeCount > 0){
					table.data = newData;
					table.save();
				}
				res.write(changeCount);
				res.end();
				log("delete", tableName, info, changeCount);
				break;
		}
	}else{
		switch(req.method){
			case "POST":
				// create table tableName columns
				readRequestData(req).then(function(info){
					var tableFile = path.join(dbPath, info.tableName + ".json");
					if(fs.existsSync(tableFile)){
						res.writeHead(412, {
							'Content-Type': 'application/json'
						});
						log("create table", "", info, "已存在");
					}else{
						fs.writeFileSync(tableFile, JSON.stringify({
							columns: info.columns,
							data: []
						}));
						res.writeHead(201, {
							'Content-Type': 'application/json'
						});
						log("create table", "", info, "成功");
					}
					res.end();
				});
				break;
			case "DELETE":
				break;
		}
	}
});

module.exports = {
	listen: function(port){
		server.listen(port);
		console.log("db server run at port " + port);
	}
}